------------------------------------------------------------------------------
--- Library with some functions for reading special tokens.
--
--- @author Frank Huch
--- @version November 2002
------------------------------------------------------------------------------

module CoosyRead where

import Maybe
import Char
import ReadShowTerm
import List (isPrefixOf)

infix 1 >>-

test1 n | n==0 = 42
        | (readInt "340")==(readInt "340") = test1 (n-1)

test2 n | n==0 = 42
        | fst(head(readsTerm "340"))+1==fst(head(readsTerm "340"))+1 = test2 (n-1)

readInt :: String -> Maybe (Int,String)
readInt str =
   let str1=(dropWhile (==' ') str) in
     if startsBy '-' str1
       then
         let numRead=readNum (dropWhile (==' ') (tail str1)) Nothing in
           if isNothing numRead
             then
               Nothing
             else
               Just (-fst (fromJust numRead),snd (fromJust numRead))
       else
         readNum str1 Nothing
 where
  readNum :: String -> Maybe Int -> Maybe (Int,String)
  readNum "" Nothing = Nothing
  readNum "" (Just n) = Just (n,"")
  readNum (c:cs) (Just n)
     | isDigit c = readNum cs (Just (10*n+readDigit c))
     | otherwise = Just (n,c:cs)
  readNum (c:cs) Nothing
     | isDigit c = readNum cs (Just (readDigit c))
     | otherwise = Nothing

readDigit :: Char -> Int
readDigit c = (ord c) - 48

readTuple0 :: String -> Maybe ((),String)
readTuple0 str = let str' = dropWhile isSpace str in
                   if startsBy '(' str'
                     then
                       let str1'=dropWhile isSpace (tail str') in
                         if startsBy ')' str1'
                           then
                             Just ((),tail str1')
                           else
                             Nothing
                     else
                       Nothing

readTuple2 :: (String -> Maybe (a,String)) ->
              (String -> Maybe (b,String)) -> 
              String
              -> Maybe ((a,b),String)
readTuple2 readA readB str =
     let str' = dropWhile isSpace str in
       if startsBy '(' str'
         then
           (readA (tail str')) >>- \ (x,str1) -> 
           let str1'=dropWhile isSpace str1 in
              if startsBy ',' str1'
                then
                  (readB (tail str1')) >>- \ (y,str2) ->
                  let str2'=dropWhile isSpace str2 in
                     if startsBy ')' str2'
                       then
                         Just ((x,y),tail str2')
                       else
                         Nothing            
                else
                  Nothing
         else
           Nothing

readMaybe :: (String -> Maybe (a,String)) -> String -> Maybe (Maybe a,String)
readMaybe readA str =
     let str'=dropWhile isSpace str in
       if isPrefixOf "Nothing" str'
         then
           Just (Nothing,drop 7 str')
         else
           if isPrefixOf "Just" str'
             then
               (readA (dropWhile isSpace (drop 4 str'))) >>- \ (x,str1) ->
               Just (Just x,str1)
             else
               Nothing

readList :: (String -> Maybe (a,String)) -> String -> Maybe ([a],String)
readList readA str = let str1 = dropWhile isSpace str in
                       if startsBy '[' str1
                         then
                           (readList' readA (tail str1)) >>- \ (readL,str2) ->
                           let str3 = dropWhile isSpace str2 in
                              if startsBy ']' str3 then
                                Just (readL,tail str3)
                              else
                                Nothing
                         else
                           Nothing
                                     
readList' :: (String -> Maybe (a,String)) -> String -> Maybe ([a],String)
readList' readA str =
    case str of 
      ']':_ -> Just ([],str)
      _ ->
        (readA (dropWhile isSpace str)) >>- \ (x,str1) ->
        let str1'=(dropWhile isSpace str1) in
           if startsBy ',' str1'
             then
               (readList' readA (tail str1')) >>- \ (xs,str2) ->
               Just (x:xs,str2)
             else
               Just ([x],str1)

(>>-) :: Maybe a -> (a -> Maybe b) -> Maybe b
Nothing >>- _ = Nothing
(Just x) >>- f  = f x

startsBy :: a -> [a] -> Bool
startsBy _ [] =False
startsBy e (e':_) = e==e'

--- Read a hexadecimal number in a string.
--- The string might contain leadings blanks and the the integer is read
--- up to the first non-heaxdecimal digit.
readHex :: String -> Int   -- result >= 0
readHex l = readHexPrefix (dropWhile (\c->c==' ') l) 0
 where
  readHexPrefix [] n = n
  readHexPrefix (c:cs) n =
   let cv = hex2int c in
     if cv>=0 then readHexPrefix cs (n*16+cv)
              else n

  hex2int c = if isDigit c then ord c - ord '0'
                           else if ord c >= ord 'A' && ord c <= ord 'F'
                                then ord c - ord 'A' + 10
                                else -1

-- end of library MyRead